"""Test IP-MAC data structure"""
import unittest

from node_ip_resolver import NodeIpResolver

class TestNodeIpResolver(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_01(self):
        """Input an entry."""
        resolver = NodeIpResolver()
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1'}
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)

    def test_02(self):
        """Input two entries with duplicate IPs.  This IP is hidden from results."""
        resolver = NodeIpResolver()
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1',
                   '00:01:02:0a:0b:0d': '192.0.2.1',
                   '0a:01:02:0a:0b:01': '192.0.2.2'}
        r = resolver.set('source1', node_ip)
        self.assertEqual({'00:01:02:0a:0b:0c': None, '00:01:02:0a:0b:0d': None, '0a:01:02:0a:0b:01': '192.0.2.2'}, r)
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1',
                   '00:01:02:0a:0b:0d': '192.0.2.3',
                   '0a:01:02:0a:0b:01': '192.0.2.2'}
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)

    def test_03(self):
        """Input two IP entries that agree, from different sources"""
        resolver = NodeIpResolver()
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1'}
        r = resolver.set('source1', node_ip)
        r = resolver.set('source2', node_ip)
        self.assertEqual(node_ip, r)

    def test_04(self):
        """A source is removed."""
        resolver = NodeIpResolver()
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1'}
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)
        r = resolver.set('source1', {})
        self.assertEqual({}, r)
        self.assertEqual(resolver.source_node_ip, {})
        self.assertEqual(resolver.ip_source_id, {})

    def test_05(self):
        """An invalid IP address is handled properly."""
        resolver = NodeIpResolver()
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.1'}
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)
        node_ip = {'00:01:02:0a:0b:0c': '169.254.0.100'}  # link-local IP
        r = resolver.set('source1', node_ip)
        self.assertEqual({'00:01:02:0a:0b:0c': None}, r)
        self.assertEqual(resolver.source_node_ip, {})
        self.assertEqual(resolver.ip_source_id, {})
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.2'}  # link-local IP
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.3'}  # invalid ip
        r = resolver.set('source1', node_ip)
        self.assertEqual(node_ip, r)
        node_ip = {'00:01:02:0a:0b:0c': '192.0.2.256'}  # invalid ip
        r = resolver.set('source1', node_ip)
        self.assertEqual({'00:01:02:0a:0b:0c': None}, r)
        self.assertEqual(resolver.source_node_ip, {})
        self.assertEqual(resolver.ip_source_id, {})



import os.path
import sys
from tabulate import tabulate

sys.path.append(os.path.dirname(__file__))  # let code under test load stubs
from device_test import DeviceTestCase

class TestRadioStats(DeviceTestCase):
    def test_trango_m900s_M900S_1p2a4H8005D04071403_01(self):
        obj = self.run_data("trango.m900s.M900S.1p2a4H8005D04071403.01")
        exp_info = {
                'model': 'M900S',
                'rf_freq': 918,
                'rf_tx_power': 26,
                'sta_count': 0,
                'sw_version': '1p2a4H8005D04071403',
                'vendor': 'Trango'
        }
        exp_stats = {
                'dl_en_bytes': 4749346,
                'dl_rf_bytes': None,
                'dl_rf_packets': 51393,
                'dl_rf_packets_dropped': None,
                'ul_en_bytes': 566932,
                'ul_rf_bytes': None,
                'ul_rf_packets': 0,
                'ul_rf_packets_dropped': 0
        }
        exp_sta_stats = {
         
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'in_octets': '4749346',
                                                              'name': 'Ethernet '
                                                                      'Interface',
                                                              'oper_up': True,
                                                              'out_octets': '566932',
                                                              'poller_hash': 'AB3hVegv'},
                                                      {       'admin_up': True,
                                                              'in_discards': '0',
                                                              'in_ucast_pkts': '0',
                                                              'mac_address': '00:01:de:15:5e:82',
                                                              'name': 'Radio Interface',
                                                              'oper_up': True,
                                                              'out_ucast_pkts': '51393',
                                                              'radios': [       {       'frequency': 918,
                                                                                        'id': '1',
                                                                                        'streams': [       {       'id': '1'}],
                                                                                        'tx_power': 26}],
                                                              'sectors': [       {       'mode': 'AP'}]}],
                                'manufacturer': 'Trango',
                                'model': 'M900S',
                                'sw_version': '1p2a4H8005D04071403',
                                'system_mac_address': '00:01:de:15:5e:82',
                                'uptime': '40109'}
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'cpe', obj.model) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in obj.stations}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))

    def test_trango_m900s_M900S_2p0r3H8005D06013101_01(self):
        obj = self.run_data("trango.m900s.M900S.2p0r3H8005D06013101.01")
        exp_info = {
                'model': 'M900S',
                'rf_freq': 924,
                'rf_tx_power': 26,
                'sta_count': 2,
                'sw_version': '2p0r3H8005D06013101',
                'vendor': 'Trango'
        }
        exp_stats = {
                'dl_en_bytes': 4185746079,
                'dl_rf_bytes': 2695062413,
                'dl_rf_packets': 192544441,
                'dl_rf_packets_dropped': None,
                'ul_en_bytes': 2418392404,
                'ul_rf_bytes': 2393603038,
                'ul_rf_packets': 142160298,
                'ul_rf_packets_dropped': 582409
        }
        exp_sta_stats = {
                '00:01:de:4c:d0:2a': {       'dev_macs': None,
                                             'dl_link_rate': 3000000,
                                             'dl_rf_bytes': 1439373111,
                                             'dl_rf_packets': 4188010,
                                             'dl_rf_packets_dropped': 67882,
                                             'dl_rssi': -84.0,
                                             'ip_address': '192.168.240.26',
                                             'mac_address': '00:01:de:4c:d0:2a',
                                             'name': 'James,Jack',
                                             'ul_link_rate': 3000000,
                                             'ul_rf_bytes': 268660059,
                                             'ul_rf_packets': 2483814,
                                             'ul_rf_packets_dropped': 89272,
                                             'ul_rssi': -82.0},
                '00:01:de:4c:db:4a': {       'dev_macs': None,
                                             'dl_link_rate': 3000000,
                                             'dl_rf_bytes': 3962643147,
                                             'dl_rf_packets': 442705837,
                                             'dl_rf_packets_dropped': 127605,
                                             'dl_rssi': -76.0,
                                             'ip_address': '192.168.240.66',
                                             'mac_address': '00:01:de:4c:db:4a',
                                             'name': 'PinestoneCrossing',
                                             'ul_link_rate': 3000000,
                                             'ul_rf_bytes': 2233665158,
                                             'ul_rf_packets': 329505234,
                                             'ul_rf_packets_dropped': 10808,
                                             'ul_rssi': -75.0}
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'in_octets': '4185746079',
                                                              'name': 'Ethernet '
                                                                      'Interface',
                                                              'oper_up': True,
                                                              'out_octets': '2418392404',
                                                              'poller_hash': 'AB3hVfe/'},
                                                      {       'admin_up': True,
                                                              'in_discards': '582409',
                                                              'in_octets': '2393603038',
                                                              'in_ucast_pkts': '142160298',
                                                              'links': [       {       'in_discards': '89272',
                                                                                       'in_octets': '268660059',
                                                                                       'in_ucast_pkts': '2483814',
                                                                                       'mac_address': '00:01:de:4c:d0:2a',
                                                                                       'out_discards': '67882',
                                                                                       'out_octets': '1439373111',
                                                                                       'out_retransmits': '384214',
                                                                                       'out_ucast_pkts': '4188010'},
                                                                               {       'in_discards': '10808',
                                                                                       'in_octets': '2233665158',
                                                                                       'in_ucast_pkts': '329505234',
                                                                                       'mac_address': '00:01:de:4c:db:4a',
                                                                                       'out_discards': '127605',
                                                                                       'out_octets': '3962643147',
                                                                                       'out_retransmits': '137101',
                                                                                       'out_ucast_pkts': '442705837'}],
                                                              'mac_address': '00:01:de:15:5f:7b',
                                                              'name': 'Radio Interface',
                                                              'oper_up': True,
                                                              'out_octets': '2695062413',
                                                              'out_ucast_pkts': '192544441',
                                                              'radios': [       {       'frequency': 924,
                                                                                        'id': '1',
                                                                                        'streams': [       {       'id': '1',
                                                                                                                   'links': [       {       'distance': 643,
                                                                                                                                            'mac_address': '00:01:de:4c:d0:2a',
                                                                                                                                            'rssi': -82.0,
                                                                                                                                            'rx_link_rate': '3000000',
                                                                                                                                            'tx_link_rate': '3000000'},
                                                                                                                                    {       'distance': 2092,
                                                                                                                                            'mac_address': '00:01:de:4c:db:4a',
                                                                                                                                            'rssi': -75.0,
                                                                                                                                            'rx_link_rate': '3000000',
                                                                                                                                            'tx_link_rate': '3000000'}]}],
                                                                                        'tx_power': 26}],
                                                              'sectors': [       {       'mode': 'AP'}]}],
                                'manufacturer': 'Trango',
                                'model': 'M900S',
                                'sw_version': '2p0r3H8005D06013101',
                                'system_mac_address': '00:01:de:15:5f:7b',
                                'uptime': '2414874'},
                'peers': [       {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '1439373111',
                                                                       'in_ucast_pkts': '4188010',
                                                                       'links': [       {       'in_octets': '1439373111',
                                                                                                'in_ucast_pkts': '4188010',
                                                                                                'mac_address': '00:01:de:15:5f:7b',
                                                                                                'out_octets': '268660059',
                                                                                                'out_ucast_pkts': '2483814'}],
                                                                       'mac_address': '00:01:de:4c:d0:2a',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '268660059',
                                                                       'out_ucast_pkts': '2483814',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '3000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 643,
                                                                                                                                                     'mac_address': '00:01:de:15:5f:7b',
                                                                                                                                                     'rssi': -84.0,
                                                                                                                                                     'rx_link_rate': '3000000',
                                                                                                                                                     'tx_link_rate': '3000000'}]}],
                                                                                                 'tx_link_rate': '3000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.240.26',
                                         'manufacturer': 'Trango',
                                         'name': 'James,Jack',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:d0:2a'},
                                 {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '3962643147',
                                                                       'in_ucast_pkts': '442705837',
                                                                       'links': [       {       'in_octets': '3962643147',
                                                                                                'in_ucast_pkts': '442705837',
                                                                                                'mac_address': '00:01:de:15:5f:7b',
                                                                                                'out_octets': '2233665158',
                                                                                                'out_ucast_pkts': '329505234'}],
                                                                       'mac_address': '00:01:de:4c:db:4a',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '2233665158',
                                                                       'out_ucast_pkts': '329505234',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '3000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 2092,
                                                                                                                                                     'mac_address': '00:01:de:15:5f:7b',
                                                                                                                                                     'rssi': -76.0,
                                                                                                                                                     'rx_link_rate': '3000000',
                                                                                                                                                     'tx_link_rate': '3000000'}]}],
                                                                                                 'tx_link_rate': '3000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.240.66',
                                         'manufacturer': 'Trango',
                                         'name': 'PinestoneCrossing',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:db:4a'}]
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'cpe', obj.model) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in obj.stations}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))

    def test_trango_m900s_M900S_2p1r1H8005D06101801_01(self):
        obj = self.run_data("trango.m900s.M900S.2p1r1H8005D06101801.01")
        exp_info = {
                'model': 'M900S',
                'rf_freq': 924,
                'rf_tx_power': 26,
                'sta_count': 4,
                'sw_version': '2p1r1H8005D06101801',
                'vendor': 'Trango'
        }
        exp_stats = {
                'dl_en_bytes': 2690227221,
                'dl_rf_bytes': 704517519,
                'dl_rf_packets': 208783723,
                'dl_rf_packets_dropped': None,
                'ul_en_bytes': 1986156428,
                'ul_rf_bytes': 722089994,
                'ul_rf_packets': 159055534,
                'ul_rf_packets_dropped': 101151371
        }
        exp_sta_stats = {
                '00:01:de:4c:d6:7a': {       'dev_macs': None,
                                             'dl_link_rate': 3000000,
                                             'dl_rf_bytes': 3001249332,
                                             'dl_rf_packets': 5781339,
                                             'dl_rf_packets_dropped': 2654,
                                             'dl_rssi': -56.0,
                                             'ip_address': '192.168.244.51',
                                             'mac_address': '00:01:de:4c:d6:7a',
                                             'name': 'Haarala,Lawrence',
                                             'ul_link_rate': 3000000,
                                             'ul_rf_bytes': 881874078,
                                             'ul_rf_packets': 4433718,
                                             'ul_rf_packets_dropped': 900,
                                             'ul_rssi': -61.0},
                '00:01:de:4c:d9:02': {       'dev_macs': None,
                                             'dl_link_rate': 3000000,
                                             'dl_rf_bytes': 4286239965,
                                             'dl_rf_packets': 4578156,
                                             'dl_rf_packets_dropped': 33236,
                                             'dl_rssi': -81.0,
                                             'ip_address': '192.168.244.44',
                                             'mac_address': '00:01:de:4c:d9:02',
                                             'name': 'Swartz,MaryLou',
                                             'ul_link_rate': 3000000,
                                             'ul_rf_bytes': 1154341091,
                                             'ul_rf_packets': 3315191,
                                             'ul_rf_packets_dropped': 73625,
                                             'ul_rssi': -80.0},
                '00:01:de:4c:db:9c': {       'dev_macs': None,
                                             'dl_link_rate': 3000000,
                                             'dl_rf_bytes': 1256904174,
                                             'dl_rf_packets': 5343700,
                                             'dl_rf_packets_dropped': 12057,
                                             'dl_rssi': -78.0,
                                             'ip_address': '192.168.244.18',
                                             'mac_address': '00:01:de:4c:db:9c',
                                             'name': 'Gherardini,Norm',
                                             'ul_link_rate': 3000000,
                                             'ul_rf_bytes': 579000134,
                                             'ul_rf_packets': 3501538,
                                             'ul_rf_packets_dropped': 25287,
                                             'ul_rssi': -76.0},
                '00:01:de:4c:db:a4': {       'dev_macs': None,
                                             'dl_link_rate': 1000000,
                                             'dl_rf_bytes': 188416047,
                                             'dl_rf_packets': 4145632,
                                             'dl_rf_packets_dropped': 22236,
                                             'dl_rssi': -81.0,
                                             'ip_address': '192.168.244.46',
                                             'mac_address': '00:01:de:4c:db:a4',
                                             'name': 'Nyman,Becky',
                                             'ul_link_rate': 1000000,
                                             'ul_rf_bytes': 439323667,
                                             'ul_rf_packets': 3207578,
                                             'ul_rf_packets_dropped': 29275,
                                             'ul_rssi': -78.0}
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'in_octets': '2690227221',
                                                              'name': 'Ethernet '
                                                                      'Interface',
                                                              'oper_up': True,
                                                              'out_octets': '1986156428',
                                                              'poller_hash': 'AB3hVwLv'},
                                                      {       'admin_up': True,
                                                              'in_discards': '101151371',
                                                              'in_octets': '722089994',
                                                              'in_ucast_pkts': '159055534',
                                                              'links': [       {       'in_discards': '25287',
                                                                                       'in_octets': '579000134',
                                                                                       'in_ucast_pkts': '3501538',
                                                                                       'mac_address': '00:01:de:4c:db:9c',
                                                                                       'out_discards': '12057',
                                                                                       'out_octets': '1256904174',
                                                                                       'out_retransmits': '248457',
                                                                                       'out_ucast_pkts': '5343700'},
                                                                               {       'in_discards': '73625',
                                                                                       'in_octets': '1154341091',
                                                                                       'in_ucast_pkts': '3315191',
                                                                                       'mac_address': '00:01:de:4c:d9:02',
                                                                                       'out_discards': '33236',
                                                                                       'out_octets': '4286239965',
                                                                                       'out_retransmits': '603716',
                                                                                       'out_ucast_pkts': '4578156'},
                                                                               {       'in_discards': '29275',
                                                                                       'in_octets': '439323667',
                                                                                       'in_ucast_pkts': '3207578',
                                                                                       'mac_address': '00:01:de:4c:db:a4',
                                                                                       'out_discards': '22236',
                                                                                       'out_octets': '188416047',
                                                                                       'out_retransmits': '257335',
                                                                                       'out_ucast_pkts': '4145632'},
                                                                               {       'in_discards': '900',
                                                                                       'in_octets': '881874078',
                                                                                       'in_ucast_pkts': '4433718',
                                                                                       'mac_address': '00:01:de:4c:d6:7a',
                                                                                       'out_discards': '2654',
                                                                                       'out_octets': '3001249332',
                                                                                       'out_retransmits': '13812',
                                                                                       'out_ucast_pkts': '5781339'}],
                                                              'mac_address': '00:01:de:15:70:2e',
                                                              'name': 'Radio Interface',
                                                              'oper_up': True,
                                                              'out_octets': '704517519',
                                                              'out_ucast_pkts': '208783723',
                                                              'radios': [       {       'frequency': 924,
                                                                                        'id': '1',
                                                                                        'streams': [       {       'id': '1',
                                                                                                                   'links': [       {       'distance': 2414,
                                                                                                                                            'mac_address': '00:01:de:4c:db:9c',
                                                                                                                                            'rssi': -76.0,
                                                                                                                                            'rx_link_rate': '3000000',
                                                                                                                                            'tx_link_rate': '3000000'},
                                                                                                                                    {       'distance': 2092,
                                                                                                                                            'mac_address': '00:01:de:4c:d9:02',
                                                                                                                                            'rssi': -80.0,
                                                                                                                                            'rx_link_rate': '3000000',
                                                                                                                                            'tx_link_rate': '3000000'},
                                                                                                                                    {       'distance': 2574,
                                                                                                                                            'mac_address': '00:01:de:4c:db:a4',
                                                                                                                                            'rssi': -78.0,
                                                                                                                                            'rx_link_rate': '1000000',
                                                                                                                                            'tx_link_rate': '1000000'},
                                                                                                                                    {       'distance': 482,
                                                                                                                                            'mac_address': '00:01:de:4c:d6:7a',
                                                                                                                                            'rssi': -61.0,
                                                                                                                                            'rx_link_rate': '3000000',
                                                                                                                                            'tx_link_rate': '3000000'}]}],
                                                                                        'tx_power': 26}],
                                                              'sectors': [       {       'mode': 'AP'}]}],
                                'manufacturer': 'Trango',
                                'model': 'M900S',
                                'sw_version': '2p1r1H8005D06101801',
                                'system_mac_address': '00:01:de:15:70:2e',
                                'uptime': '315123'},
                'peers': [       {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '1256904174',
                                                                       'in_ucast_pkts': '5343700',
                                                                       'links': [       {       'in_octets': '1256904174',
                                                                                                'in_ucast_pkts': '5343700',
                                                                                                'mac_address': '00:01:de:15:70:2e',
                                                                                                'out_octets': '579000134',
                                                                                                'out_ucast_pkts': '3501538'}],
                                                                       'mac_address': '00:01:de:4c:db:9c',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '579000134',
                                                                       'out_ucast_pkts': '3501538',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '3000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 2414,
                                                                                                                                                     'mac_address': '00:01:de:15:70:2e',
                                                                                                                                                     'rssi': -78.0,
                                                                                                                                                     'rx_link_rate': '3000000',
                                                                                                                                                     'tx_link_rate': '3000000'}]}],
                                                                                                 'tx_link_rate': '3000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.244.18',
                                         'manufacturer': 'Trango',
                                         'name': 'Gherardini,Norm',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:db:9c'},
                                 {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '4286239965',
                                                                       'in_ucast_pkts': '4578156',
                                                                       'links': [       {       'in_octets': '4286239965',
                                                                                                'in_ucast_pkts': '4578156',
                                                                                                'mac_address': '00:01:de:15:70:2e',
                                                                                                'out_octets': '1154341091',
                                                                                                'out_ucast_pkts': '3315191'}],
                                                                       'mac_address': '00:01:de:4c:d9:02',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '1154341091',
                                                                       'out_ucast_pkts': '3315191',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '3000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 2092,
                                                                                                                                                     'mac_address': '00:01:de:15:70:2e',
                                                                                                                                                     'rssi': -81.0,
                                                                                                                                                     'rx_link_rate': '3000000',
                                                                                                                                                     'tx_link_rate': '3000000'}]}],
                                                                                                 'tx_link_rate': '3000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.244.44',
                                         'manufacturer': 'Trango',
                                         'name': 'Swartz,MaryLou',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:d9:02'},
                                 {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '188416047',
                                                                       'in_ucast_pkts': '4145632',
                                                                       'links': [       {       'in_octets': '188416047',
                                                                                                'in_ucast_pkts': '4145632',
                                                                                                'mac_address': '00:01:de:15:70:2e',
                                                                                                'out_octets': '439323667',
                                                                                                'out_ucast_pkts': '3207578'}],
                                                                       'mac_address': '00:01:de:4c:db:a4',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '439323667',
                                                                       'out_ucast_pkts': '3207578',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '1000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 2574,
                                                                                                                                                     'mac_address': '00:01:de:15:70:2e',
                                                                                                                                                     'rssi': -81.0,
                                                                                                                                                     'rx_link_rate': '1000000',
                                                                                                                                                     'tx_link_rate': '1000000'}]}],
                                                                                                 'tx_link_rate': '1000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.244.46',
                                         'manufacturer': 'Trango',
                                         'name': 'Nyman,Becky',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:db:a4'},
                                 {       'interfaces': [       {       'admin_up': True,
                                                                       'in_octets': '3001249332',
                                                                       'in_ucast_pkts': '5781339',
                                                                       'links': [       {       'in_octets': '3001249332',
                                                                                                'in_ucast_pkts': '5781339',
                                                                                                'mac_address': '00:01:de:15:70:2e',
                                                                                                'out_octets': '881874078',
                                                                                                'out_ucast_pkts': '4433718'}],
                                                                       'mac_address': '00:01:de:4c:d6:7a',
                                                                       'name': 'Radio '
                                                                               'Interface',
                                                                       'oper_up': True,
                                                                       'out_octets': '881874078',
                                                                       'out_ucast_pkts': '4433718',
                                                                       'radios': [       {       'frequency': 924,
                                                                                                 'id': '1',
                                                                                                 'rx_link_rate': '3000000',
                                                                                                 'streams': [       {       'id': '1',
                                                                                                                            'links': [       {       'distance': 482,
                                                                                                                                                     'mac_address': '00:01:de:15:70:2e',
                                                                                                                                                     'rssi': -56.0,
                                                                                                                                                     'rx_link_rate': '3000000',
                                                                                                                                                     'tx_link_rate': '3000000'}]}],
                                                                                                 'tx_link_rate': '3000000'}],
                                                                       'sectors': [       {       'mode': 'STA'}]}],
                                         'management_ip': '192.168.244.51',
                                         'manufacturer': 'Trango',
                                         'name': 'Haarala,Lawrence',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'system_mac_address': '00:01:de:4c:d6:7a'}]
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'cpe', obj.model) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in obj.stations}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))



"""Test airtime calculations scenarios"""
import asyncio
from collections import namedtuple, Counter
import os, os.path
import sys
import time
import unittest
from uuid import UUID

from preseem import FakeNetworkMetadataModel, FakeNetworkMetricsModel, NetworkMetadataReference, NetworkMetadataEntity, Reference

sys.path.append(os.path.dirname(__file__))  # let code under test load stubs
import ap
import ap_data
from fake_snmp import FakeSnmpClient
from ne import NetworkElementRegistry
from preseem import network_element_update
from devices.ubnt import snmpBadMibError, snmpUnsupportedFwError
from device_test import get_datafiles
from preseem_protobuf.network_poller import network_poller_pb2
from preseem.network_element_update import ErrorContext, NetworkElementUpdate, NetworkElement
from fake_context import FakeHttpClient, fake_ping, FakeContext

import pprint
def pp(name, d, indent=8):
    print(' ' * indent + '{} = '.format(name) + '{\n ', end='')
    for line in pprint.pformat(d, indent=indent)[1:-1].split('\n'):
        print(' ' * indent + line)
    print(' ' * indent + '}')

class Ap(ap.Ap):
    """We subclass the Ap object to add some test synchronization helpers"""
    async def poll(self):
        FakeContext.ap_event.clear()
        await super().poll()
        FakeContext.ap_event.set()

# What I want here is a Module I can add to ap.
# Really I would rather not even do any SNMP here, we're not testing that logic
# and in theory SNMP would be optional anyway.
# I could monkey-patch ap.poll() - sort of already did inAp above..could do
# similar and just skip all the polling logic, which I think may do away
# with any snmp requirements?
# YES - then just call self.load() to do the AP algorithms, after I set
# self.module to my fake module.
#
# May not even need anything fake: just construct a couple AP objects and
# call load on them?

# Ok so try this:
# 1. Create Ap object manually instead of going through registry.
# 2. Set module on it to a FakeModule that I can control to return whatever
# stations list I want, move those around.
class FakeModule:
    def __init__(self):
        self.stations = []
        self.model = None
        self.mode = "ap"
        self.station_poller = None

    async def poll(self):
        pass

    async def poll_station(self, sta, snmp_client, http_client):
        if self.station_poller:
            return self.station_poller(sta, snmp_client, http_client)
        return sta
    

class FakeClient:
    def __init__(self, host=None):
        self.dft = {}
        self.host = host


class TestAP(unittest.TestCase):
    def setUp(self):
        apdd = ap_data.ApData('ap_info.yaml')
        self.datafiles = get_datafiles(apdd)
        self.ctx = FakeContext()
        self.loop = asyncio.new_event_loop() # needed to initialize asyncio
        self.reg = NetworkElementRegistry(self.ctx, {}, Ap)
        self.reg.ne_type = 'ap'
        self.ctx.company_uuid = UUID('4a24ad99-d502-3846-a8de-6c202c665a37')
        asyncio.set_event_loop(self.loop)
        self._await(self.ctx.start())

    def tearDown(self):
        self._await(self.reg.close())
        self._await(self.ctx.close())
        self.loop.close()

    def _await(self, co):
        return self.loop.run_until_complete(co)

    def wait_for_ap_poll(self):
        """Wait for an AP poll to complete."""
        self._await(self.ctx.ap_event.wait())


    def test_airtime_calc(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations """
        df = self.datafiles.get(('cambium-epmp.epmp.ePMP 2000.4.5.6.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        total_tx_airtime = 0
        total_rx_airtime = 0

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())
        ne_u= ap.module.network_element_update_msg
        ne_u.data.uptime += 60


        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets += 6000000
            ne_u.data.interfaces[1].links[n].out_octets += 3000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 100000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 200000000

        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].radios[0].streams[0].links)):
            link_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            link_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            rx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate)
            tx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate)

            self.assertEqual(round(link_rx_airtime,5), 6000000 * 8 * 100 / (rx_link_rate * 60))
            self.assertEqual(round(link_tx_airtime,5), 3000000 * 8 * 100 / (tx_link_rate * 60))
            total_rx_airtime += link_rx_airtime
            total_tx_airtime += link_tx_airtime

        # port rf level tx_airtime equals reported tx_frame utilization and rx_airtime equals sum of stations' airtimes for ePMP models
        self.assertEqual(ne_u.data.interfaces[1].radios[0].rx_airtime, round(total_rx_airtime,5))
        self.assertEqual(ne_u.data.interfaces[1].radios[0].tx_airtime, ne_u.data.interfaces[1].radios[0].tx_frutl)

    def test_airtime_calc2(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations in case of pre-existing SM level airtimes """
        df = self.datafiles.get(('ubnt.airmax-ac.Rocket 5AC PTMP.v8.5.0.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        total_tx_airtime = 0
        total_rx_airtime = 0

        self._await(ap.poll())
        self.wait_for_ap_poll()
        ne_u= ap.module.network_element_update_msg

        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[0].radios[0].streams[0].links)):
            link_rx_airtime = ne_u.data.interfaces[0].radios[0].streams[0].links[n].rx_airtime
            link_tx_airtime = ne_u.data.interfaces[0].radios[0].streams[0].links[n].tx_airtime
            self.assertIsNotNone(link_rx_airtime)
            self.assertIsNotNone(link_tx_airtime)
            total_rx_airtime += link_rx_airtime
            total_tx_airtime += link_tx_airtime

        self.assertEqual(round(ne_u.data.interfaces[0].radios[0].rx_airtime, 5), round(total_rx_airtime,5))
        self.assertEqual(round(ne_u.data.interfaces[0].radios[0].tx_airtime, 5),  round(total_tx_airtime,5))

    def test_airtime_calc3(self):
        """Test the airtime calculations in case no conn_time available from Stations. We use AP's uptime instead"""
        ap = Ap(self.ctx, 'TestNE', 'Test Element', 'site', None, None, 'ap', None)
        self._await(ap.log_ne_update()) # Resolve race condition with NEU message
        ap.module = FakeModule()
        ap.errctx = getattr(ap.module, 'errctx', ErrorContext())
        ap.network_element_update_msg = getattr(ap.module, 'network_element_update_msg', NetworkElementUpdate(errctx=ap.errctx))
        Sta = namedtuple('Sta', ('mac_address'))
        ap.module.stations = [Sta('08:55:31:0a:d4:38')]
        msg = ap.network_element_update_msg
        msg.data.uptime = 1000
        wlan = msg.data.interfaces.add()
        link, radio = wlan.links.add(), wlan.radios.add()
        link.in_octets = 402579957
        link.out_octets = 505139803
        link.poller_hash = b'\x01'
        stream = radio.streams.add()
        wlink = stream.links.add()
        wlink.tx_link_rate = 2310000000
        wlink.rx_link_rate = 2810000000
        wlink.poller_hash = b'\x01'

        total_tx_airtime = 0
        total_rx_airtime = 0
        ap.stations = ap.module.stations
        self._await(ap.log_ne_update())
        ne_u = ap.network_element_update_msg

        ne_u.data.uptime += 60
        for n in range(len(ne_u.data.interfaces[0].links)):
            ne_u.data.interfaces[0].links[n].in_octets += 6000000
            ne_u.data.interfaces[0].links[n].out_octets += 3000000
            ne_u.data.interfaces[0].radios[0].streams[0].links[n].rx_link_rate = 10000000
            ne_u.data.interfaces[0].radios[0].streams[0].links[n].tx_link_rate = 20000000
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[0].radios[0].streams[0].links)):
            link_rx_airtime = ne_u.data.interfaces[0].radios[0].streams[0].links[n].rx_airtime
            link_tx_airtime = ne_u.data.interfaces[0].radios[0].streams[0].links[n].tx_airtime
            rx_link_rate = float(ne_u.data.interfaces[0].radios[0].streams[0].links[n].rx_link_rate)
            tx_link_rate = float(ne_u.data.interfaces[0].radios[0].streams[0].links[n].tx_link_rate)
    
            self.assertEqual(link_rx_airtime, 6000000 * 8 * 100 / (rx_link_rate * 60))
            self.assertEqual(link_tx_airtime, 3000000 * 8 * 100 / (tx_link_rate * 60))
            total_rx_airtime += link_rx_airtime
            total_tx_airtime += link_tx_airtime

        self.assertEqual(ne_u.data.interfaces[0].radios[0].rx_airtime,  total_rx_airtime)
        self.assertEqual(ne_u.data.interfaces[0].radios[0].tx_airtime,  total_tx_airtime)

    def test_airtime_calc4(self):
        """Test airtime for wireless links in case of multiple streams """
        df = self.datafiles.get(('cambium-canopy.450.PMP 450.15.1.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)

        self._await(ap.poll())
        self.wait_for_ap_poll()
        ne_u = ap.module.network_element_update_msg
        link_len = len(ne_u.data.interfaces[1].links)

        # To populate missing rx_link_rate in .snmp for airtime_calc
        for n in range(link_len):
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 10000000
            ne_u.data.interfaces[1].radios[0].streams[1].links[n].rx_link_rate = 10000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 20000000
            ne_u.data.interfaces[1].radios[0].streams[1].links[n].tx_link_rate = 20000000
        self._await(ap.log_ne_update())

        for n in range(link_len):
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 10000000
            ne_u.data.interfaces[1].radios[0].streams[1].links[n].rx_link_rate = 10000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 20000000
            ne_u.data.interfaces[1].radios[0].streams[1].links[n].tx_link_rate = 20000000
            ne_u.data.interfaces[1].links[n].in_octets += 6000000
            ne_u.data.interfaces[1].links[n].out_octets += 3000000
        ne_u.data.uptime += 60
        self._await(ap.log_ne_update())

        for n in range(link_len):
            streamv_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            streamh_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[1].links[n].rx_airtime
            streamv_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            streamh_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[1].links[n].tx_airtime
            rx_link_rate = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate
            tx_link_rate = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate
            exp_rx_airtime = 6000000 * 8 * 100 / (rx_link_rate * 60)
            exp_tx_airtime = 3000000 * 8 * 100 / (tx_link_rate * 60)
            self.assertEqual(streamv_rx_airtime, exp_rx_airtime)
            self.assertEqual(streamh_rx_airtime, exp_rx_airtime)
            self.assertEqual(streamv_tx_airtime, exp_tx_airtime)
            self.assertEqual(streamh_tx_airtime, exp_tx_airtime)
            self.assertEqual(streamv_rx_airtime,streamh_rx_airtime)
            self.assertEqual(streamv_tx_airtime,streamh_tx_airtime)
        
        # port rf level airtime equals reported frame utilization for Canopy 450 (both directions) nop multiplexing gain is available
        self.assertEqual(ne_u.data.interfaces[1].radios[0].tx_airtime,  ne_u.data.interfaces[1].radios[0].tx_frutl)
        self.assertEqual(ne_u.data.interfaces[1].radios[0].rx_airtime,  ne_u.data.interfaces[1].radios[0].rx_frutl)

    def test_airtime_calc5(self):
        """Test airtime for directly reported Frutl multiplied by Gain """
        df = self.datafiles.get(('cambium-canopy.450m.PMP 450m.CANOPY 20.3.1 AP.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())
        ne_u = ap.module.network_element_update_msg
   
        # port rf level airtime equals reported frame utilization for Canopy 450m with addition of mutiplexing gain (both directions)
        self.assertEqual(round(ne_u.data.interfaces[1].radios[0].tx_airtime,5),  round(ne_u.data.interfaces[1].radios[0].tx_frutl*ne_u.data.interfaces[1].radios[0].tx_multiplexing_gain, 5))
        self.assertEqual(round(ne_u.data.interfaces[1].radios[0].rx_airtime, 5),  round(ne_u.data.interfaces[1].radios[0].rx_frutl*ne_u.data.interfaces[1].radios[0].rx_multiplexing_gain, 5))

    def test_airtime_calc6(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations if one of station was reseted/rebooted """
        df = self.datafiles.get(('cambium-epmp.epmp.ePMP 2000.4.5.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        total_tx_airtime = 0
        total_rx_airtime = 0

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())
        ne_u= ap.module.network_element_update_msg


        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets += 6000000
            ne_u.data.interfaces[1].links[n].out_octets += 3000000
            if n==0:
                ne_u.data.interfaces[1].links[0].in_octets = 1 # set some low value
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 100000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 200000000
        ne_u.data.uptime += 60
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].radios[0].streams[0].links)):
            link_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            link_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            rx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate)
            tx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate)

            self.assertEqual(round(link_tx_airtime,5), 3000000 * 8 * 100 / (tx_link_rate * 60))
            if n==0:
                self.assertIsNone(link_rx_airtime)
            else:
                self.assertEqual(round(link_rx_airtime,5), 6000000 * 8 * 100 / (rx_link_rate * 60))
            total_tx_airtime += link_tx_airtime
            if link_rx_airtime is not None:
                total_rx_airtime += link_rx_airtime

        # port rf level tx_airtime equals reported tx_frame utilization and rx_airtime equals sum of stations' airtimes for ePMP models
        self.assertEqual(ne_u.data.interfaces[1].radios[0].rx_airtime,  round(total_rx_airtime,5))
        self.assertEqual(ne_u.data.interfaces[1].radios[0].tx_airtime,  ne_u.data.interfaces[1].radios[0].tx_frutl)


    def test_airtime_calc7(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations == None , No Mikrotik API """
        df = self.datafiles.get(('mikrotik.mikrotik-ap.RB411.6.25.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        total_tx_airtime = 0
        total_rx_airtime = 0

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())
        ne_u= ap.module.network_element_update_msg


        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets += 6000000
            ne_u.data.interfaces[1].links[n].out_octets += 3000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 20000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 10000000
        ne_u.data.uptime += 60
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].radios[0].streams[0].links)):
            link_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            tx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate)
            link_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            rx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate)
            self.assertIsNone(link_rx_airtime) 
            self.assertIsNone(link_tx_airtime)

        # port rf level airtime equals sum of stations' airtime for Mikrotik models (both directions) is None 
        self.assertIsNone(ne_u.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNone(ne_u.data.interfaces[1].radios[0].tx_airtime)


    def test_airtime_calc8(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations == None , No Mikrotik API """
        df = self.datafiles.get(('mikrotik.mikrotik-ap.RB411.6.25.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        total_tx_airtime = 0
        total_rx_airtime = 0
        ap.in_out_bytes = {}
        

        self._await(ap.poll())
        ne_u= ap.module.network_element_update_msg
        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets = 0
            ne_u.data.interfaces[1].links[n].out_octets = 0

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets += 0
            ne_u.data.interfaces[1].links[n].out_octets += 0
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = 20000000
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = 10000000
        ne_u.data.uptime += 60
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].radios[0].streams[0].links)):
            link_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            tx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate)
            link_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            rx_link_rate = float(ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate)
            self.assertIsNone(link_rx_airtime)
            self.assertIsNone(link_tx_airtime)


        # port rf level airtime equals sum of stations' airtime for Mikrotik models (both directions)
        self.assertIsNone(ne_u.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNone(ne_u.data.interfaces[1].radios[0].tx_airtime)


    def test_airtime_calc9(self):
        """Test the creation of network_element_update_msg and correctness of airtime calculations None if no link_rate"""
        df = self.datafiles.get(('mikrotik.mikrotik-ap.RB411.6.25.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        ap.in_out_bytes = {}
        

        self._await(ap.poll())
        ne_u= ap.module.network_element_update_msg
        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets = 0
            ne_u.data.interfaces[1].links[n].out_octets = 0
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = None
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = None

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].links)):
            ne_u.data.interfaces[1].links[n].in_octets += 0
            ne_u.data.interfaces[1].links[n].out_octets += 0
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_link_rate = None
            ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_link_rate = None
        ne_u.data.uptime += 60
        self._await(ap.log_ne_update())

        for n in range(len(ne_u.data.interfaces[1].radios[0].streams[0].links)):
            link_tx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].tx_airtime
            link_rx_airtime = ne_u.data.interfaces[1].radios[0].streams[0].links[n].rx_airtime
            self.assertEqual(link_rx_airtime, None)
            self.assertEqual(link_tx_airtime, None)

        # port rf level airtime equals sum of stations' airtime for Mikrotik models (both directions)
        self.assertEqual(ne_u.data.interfaces[1].radios[0].rx_airtime, None)
        self.assertEqual(ne_u.data.interfaces[1].radios[0].tx_airtime, None)


    def test_airtime_calc10(self):
        """Test the airtime calculations when the peers have the same AP counter_source"""
        df = self.datafiles.get(('cambium-epmp.epmp3000.ePMP 3000.4.3.2.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        self._await(ap.log_ne_update())
        
        # Poll 1: Set to AP counter_source
        self._await(ap.poll())
        msg = ap.module.network_element_update_msg
        msg.peers[0].counter_source = "AP"

        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets = 1000
            msg.data.interfaces[1].links[n].out_octets = 1000

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        msg.data.interfaces[1].radios[0].streams[0].links[0].connected_time += 60 
        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets += 6000000
            msg.data.interfaces[1].links[n].out_octets += 3000000

        # Poll 2: Keep AP as counter_source
        msg.data.uptime += 60
        msg.peers[0].counter_source = "AP"
        self._await(ap.log_ne_update())

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        self.assertIsNotNone(msg.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNotNone(msg.data.interfaces[1].radios[0].tx_airtime)


    def test_airtime_calc11(self):
        """Test the airtime calculations when the peers have the same CPE counter_source"""
        df = self.datafiles.get(('cambium-epmp.epmp3000.ePMP 3000.4.3.2.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        self._await(ap.log_ne_update())
        
        # Poll 1: Set to CPE counter_source
        self._await(ap.poll())
        msg = ap.module.network_element_update_msg
        msg.peers[0].counter_source = "CPE"

        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets = 1000
            msg.data.interfaces[1].links[n].out_octets = 1000

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        msg.data.interfaces[1].radios[0].streams[0].links[0].connected_time += 60 
        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets += 6000000
            msg.data.interfaces[1].links[n].out_octets += 3000000

        # Poll 2: Keep as CPE counter_source
        msg.data.uptime += 60
        msg.peers[0].counter_source = "CPE"
        self._await(ap.log_ne_update())

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        self.assertIsNotNone(msg.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNotNone(msg.data.interfaces[1].radios[0].tx_airtime)


    def test_airtime_calc12(self):
        """Test the airtime calculations when the peers AP and CPE have different counter_source"""
        df = self.datafiles.get(('cambium-epmp.epmp3000.ePMP 3000.4.3.2.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        self._await(ap.log_ne_update())
        
        # Poll 1: Set to AP counter_source
        self._await(ap.poll())
        msg = ap.module.network_element_update_msg
        msg.peers[0].counter_source = "AP"

        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets = 1000
            msg.data.interfaces[1].links[n].out_octets = 1000

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        msg.data.interfaces[1].radios[0].streams[0].links[0].connected_time += 60 
        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets += 6000000
            msg.data.interfaces[1].links[n].out_octets += 3000000

        # Poll 2: Change to CPE counter_source
        msg.data.uptime += 60
        msg.peers[0].counter_source = "CPE"
        self._await(ap.log_ne_update())

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        self.assertIsNone(msg.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNone(msg.data.interfaces[1].radios[0].tx_airtime)


    def test_airtime_calc13(self):
        """Test the airtime calculations when the peers CPE and AP have different counter_source"""
        df = self.datafiles.get(('cambium-epmp.epmp3000.ePMP 3000.4.3.2.01', 'snmp'))
        ne = self._await(self.reg.set('TestNE', {'name': 'Test Element', 'site': 'My Site', 'path': df.path}))
        ap = Ap(self.ctx, ne.id, ne.name, ne.site, ne.snmp_ne, ne.snmp_ne_v3, ne.ne_type, ne.holdoff)
        self._await(ap.log_ne_update())
        
        # Poll 1: Set to CPE counter_source
        self._await(ap.poll())
        msg = ap.module.network_element_update_msg
        msg.peers[0].counter_source = "CPE"

        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets = 1000
            msg.data.interfaces[1].links[n].out_octets = 1000

        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        msg.data.interfaces[1].radios[0].streams[0].links[0].connected_time += 60 
        for n in range(len(msg.data.interfaces[1].links)):
            msg.data.interfaces[1].links[n].in_octets += 6000000
            msg.data.interfaces[1].links[n].out_octets += 3000000

        # Poll 2: Change to AP counter_source
        msg.data.uptime += 60
        msg.peers[0].counter_source = "AP"
        self._await(ap.log_ne_update())

        self._await(ap.poll())
        self.wait_for_ap_poll()
        self._await(ap.log_ne_update())

        self.assertIsNone(msg.data.interfaces[1].radios[0].rx_airtime)
        self.assertIsNone(msg.data.interfaces[1].radios[0].tx_airtime)
